@@ -1,4 +1,4 @@
-/* Copyright (C) 2006, 2007, 2008, 2009, 2012, 2013, 2014, 2016, 2018, 2019 Thorsten Kukuk
+/* Copyright (C) 2006-2021 Thorsten Kukuk
    Author: Thorsten Kukuk <kukuk@thkukuk.de>
 
    This program is free software; you can redistribute it and/or modify
@@ -247,7 +247,7 @@ relink (const char *sysconfdir, const char *file, const char *file_pc)
 static int
 replace_obsolete_modules (pam_module_t **module_list)
 {
-  int with_unix2, with_pwcheck;
+  int with_unix2, with_pwcheck, with_cracklib;
 
   with_unix2 = is_module_enabled (module_list, "pam_unix2.so", ACCOUNT) |
     is_module_enabled (module_list, "pam_unix2.so", AUTH) |
@@ -259,7 +259,9 @@ replace_obsolete_modules (pam_module_t **module_list)
     is_module_enabled (module_list, "pam_pwcheck.so", PASSWORD) |
     is_module_enabled (module_list, "pam_pwcheck.so", SESSION);
 
-  if (with_unix2 && check_for_pam_module ("pam_unix2.so", 0) == 1)
+  with_cracklib = is_module_enabled (module_list, "pam_cracklib.so", PASSWORD);
+
+  if (with_unix2 && check_for_pam_module ("pam_unix2.so", 0, 1) == 1)
     {
       option_set_t *opt_set, *opt_set2;
       pam_module_t *pam_unix2 = lookup (module_list, "pam_unix2.so");
@@ -326,22 +328,60 @@ replace_obsolete_modules (pam_module_t **module_list)
 	return 1;
     }
 
-  if (with_pwcheck && check_for_pam_module ("pam_pwcheck.so", 0) == 1)
+  /* If pam_cracklib is enabled and pam_pwquality.so installed, replace it
+     with pam_pwquality. */
+  if (with_cracklib && check_for_pam_module ("pam_pwquality.so", 0, 1) == 0)
     {
-      option_set_t *opt_cracklib, *opt_pwcheck, *opt_pwhistory;
-      pam_module_t *pam_pwcheck = lookup (module_list, "pam_pwcheck.so");
+      char *cracklib_opts_bool[] = {"debug", "reject_username",
+				    "gecoscheck", "enforce_for_root"};
+      char *cracklib_opts_str[] = {"authtok_type", "retry", "difok",
+				   "minlen", "dcredit",
+				   "ucredit", "lcredit", "ocredit",
+				   "minclass", "dictpath", "maxrepeat",
+				   "maxsequence", "maxclassrepeat"};
+      option_set_t *opt_cracklib, *opt_pwquality;
       pam_module_t *pam_cracklib = lookup (module_list, "pam_cracklib.so");
+      pam_module_t *pam_pwquality = lookup (module_list, "pam_pwquality.so");
+
+      fprintf (stderr, "pam_cracklib used, replacing with pam_pwquality.\n");
+
+      opt_cracklib = pam_cracklib->get_opt_set (pam_cracklib, PASSWORD);
+      opt_pwquality = pam_pwquality->get_opt_set (pam_pwquality, PASSWORD);
+
+      opt_cracklib->enable (opt_cracklib, "is_enabled", FALSE);
+      opt_pwquality->enable (opt_pwquality, "is_enabled", TRUE);
+
+      for (size_t i = 0; i < sizeof (cracklib_opts_bool)/sizeof (char *); i++)
+
+	opt_pwquality->enable (opt_pwquality, cracklib_opts_bool[i],
+				opt_cracklib->is_enabled (opt_cracklib,
+							  cracklib_opts_bool[i]));
+      for (size_t i = 0; i < sizeof (cracklib_opts_str)/sizeof (char *); i++)
+
+	opt_pwquality->set_opt (opt_pwquality, cracklib_opts_str[i],
+				opt_cracklib->get_opt (opt_cracklib,
+						       cracklib_opts_str[i]));
+
+      if (sanitize_check_password (module_list, 0) != 0)
+	return 1;
+    }
+
+  if (with_pwcheck && check_for_pam_module ("pam_pwcheck.so", 0, 1) == 1)
+    {
+      option_set_t *opt_pwquality, *opt_pwcheck, *opt_pwhistory;
+      pam_module_t *pam_pwcheck = lookup (module_list, "pam_pwcheck.so");
+      pam_module_t *pam_pwquality = lookup (module_list, "pam_pwquality.so");
       pam_module_t *pam_pwhistory = lookup (module_list, "pam_pwhistory.so");
 
-      fprintf (stderr, "pam_pwcheck used but not installed, replacing with pam_cracklib.\n");
+      fprintf (stderr, "pam_pwcheck used but not installed, replacing with pam_pwquality.\n");
 
       opt_pwhistory = pam_pwhistory->get_opt_set (pam_pwhistory, PASSWORD);
-      opt_cracklib = pam_cracklib->get_opt_set (pam_cracklib, PASSWORD);
-      opt_cracklib->enable (opt_cracklib, "is_enabled", TRUE);
+      opt_pwquality = pam_pwquality->get_opt_set (pam_pwquality, PASSWORD);
+      opt_pwquality->enable (opt_pwquality, "is_enabled", TRUE);
       opt_pwcheck = pam_pwcheck->get_opt_set (pam_pwcheck, PASSWORD);
       opt_pwcheck->enable (opt_pwcheck, "is_enabled", FALSE);
       if (opt_pwcheck->is_enabled (opt_pwcheck, "debug"))
-	opt_cracklib->enable (opt_cracklib, "debug", TRUE);
+	opt_pwquality->enable (opt_pwquality, "debug", TRUE);
 
       /* check, if we need to enable pam_pwhistory */
       if (opt_pwcheck->get_opt (opt_pwcheck, "remember") != NULL &&
@@ -502,6 +542,16 @@ main (int argc, char *argv[])
 	goto load_old_config_error;
       if (load_config (confdir, CONF_SESSION, SESSION, common_module_list, 1) != 0)
 	goto load_old_config_error;
+      if (load_config (confdir, CONF_SESSION_NONLOGIN, SESSION, common_module_list, 1) != 0)
+	goto load_old_config_error;
+      if (load_config (confdir, CONF_POSTLOGIN_ACCOUNT, ACCOUNT, common_module_list, 1) != 0)
+	goto load_old_config_error;
+      if (load_config (confdir, CONF_POSTLOGIN_AUTH, AUTH, common_module_list, 1) != 0)
+	goto load_old_config_error;
+      if (load_config (confdir, CONF_POSTLOGIN_PASSWORD, PASSWORD, common_module_list, 1) != 0)
+	goto load_old_config_error;
+      if (load_config (confdir, CONF_POSTLOGIN_SESSION, SESSION, common_module_list, 1) != 0)
+	goto load_old_config_error;
     }
   else if (strcmp (argv[1], "--update") == 0)
     {
@@ -540,16 +590,76 @@ main (int argc, char *argv[])
 	{
 	  if (load_config (confdir, CONF_ACCOUNT_PC, ACCOUNT, common_module_list, 1) != 0)
 	    {
-	    load_config_error:
-	      fprintf (stderr, _("\nCouldn't load config file, aborted!\n"));
-	      return 1;
+	      if (load_config (confdir, CONF_ACCOUNT, ACCOUNT, common_module_list, 1) != 0)
+		{
+		  fprintf (stderr, _("\nCouldn't load '%s', aborted!\n"), CONF_ACCOUNT);
+		  return 1;
+		}
 	    }
 	  if (load_config (confdir, CONF_AUTH_PC, AUTH, common_module_list, 1) != 0)
-	    goto load_config_error;
+	    {
+	      if (load_config (confdir, CONF_AUTH, AUTH, common_module_list, 1) != 0)
+		{
+		  fprintf (stderr, _("\nCouldn't load '%s', aborted!\n"), CONF_AUTH);
+		  return 1;
+		}
+	    }
 	  if (load_config (confdir, CONF_PASSWORD_PC, PASSWORD, common_module_list, 1) != 0)
-	    goto load_config_error;
+	    {
+	      if (load_config (confdir, CONF_PASSWORD, PASSWORD, common_module_list, 1) != 0)
+		{
+		  fprintf (stderr, _("\nCouldn't load '%s', aborted!\n"), CONF_PASSWORD);
+		  return 1;
+		}
+	    }
 	  if (load_config (confdir, CONF_SESSION_PC, SESSION, common_module_list, 1) != 0)
-	    goto load_config_error;
+	    {
+	      if (load_config (confdir, CONF_SESSION, SESSION, common_module_list, 1) != 0)
+		{
+		  fprintf (stderr, _("\nCouldn't load '%s', aborted!\n"), CONF_SESSION);
+		  return 1;
+		}
+	    }
+	  if (load_config (confdir, CONF_SESSION_NONLOGIN_PC, SESSION, common_module_list, 1) != 0)
+	    {
+	      if (load_config (confdir, CONF_SESSION_NONLOGIN, SESSION, common_module_list, 1) != 0)
+		{
+		  fprintf (stderr, _("\nCouldn't load '%s', aborted!\n"), CONF_SESSION_NONLOGIN);
+		  return 1;
+		}
+	    }
+	  if (load_config (confdir, CONF_POSTLOGIN_ACCOUNT_PC, ACCOUNT, common_module_list, 1) != 0)
+	    {
+	      if (load_config (confdir, CONF_POSTLOGIN_ACCOUNT, ACCOUNT, common_module_list, 1) != 0)
+		{
+		  fprintf (stderr, _("\nCouldn't load '%s', aborted!\n"), CONF_POSTLOGIN_ACCOUNT);
+		  return 1;
+		}
+	    }
+	  if (load_config (confdir, CONF_POSTLOGIN_AUTH_PC, AUTH, common_module_list, 1) != 0)
+	    {
+	      if (load_config (confdir, CONF_POSTLOGIN_AUTH, AUTH, common_module_list, 1) != 0)
+		{
+		  fprintf (stderr, _("\nCouldn't load '%s', aborted!\n"), CONF_POSTLOGIN_AUTH);
+		  return 1;
+		}
+	    }
+	  if (load_config (confdir, CONF_POSTLOGIN_PASSWORD_PC, PASSWORD, common_module_list, 1) != 0)
+	    {
+	      if (load_config (confdir, CONF_POSTLOGIN_PASSWORD, PASSWORD, common_module_list, 1) != 0)
+		{
+		  fprintf (stderr, _("\nCouldn't load '%s', aborted!\n"), CONF_POSTLOGIN_PASSWORD);
+		  return 1;
+		}
+	    }
+	  if (load_config (confdir, CONF_POSTLOGIN_SESSION_PC, SESSION, common_module_list, 1) != 0)
+	    {
+	      if (load_config (confdir, CONF_POSTLOGIN_SESSION, SESSION, common_module_list, 1) != 0)
+		{
+		  fprintf (stderr, _("\nCouldn't load '%s', aborted!\n"), CONF_POSTLOGIN_SESSION);
+		  return 1;
+		}
+	    }
 	}
       else
 	{
@@ -558,8 +668,7 @@ main (int argc, char *argv[])
 	    {
 	    load_config_error2:
 	      fprintf (stderr,
-		       _("\nCouldn't load config file '%s/pam.d/%s', aborted!\n"),
-		       confdir, gl_service);
+		       _("\nCouldn't load config file '%s', aborted!\n"), gl_service);
 	      return 1;
 	    }
 	  if (load_config (confdir, gl_service, AUTH, service_module_list, 0) != 0)
@@ -590,7 +699,6 @@ main (int argc, char *argv[])
         {"winbind",               no_argument,       NULL, 2200 },
         {"winbind-debug",         no_argument,       NULL, 2201 },
 	{"sss",                   no_argument,       NULL, 2300 },
-	{"ecryptfs",              no_argument,       NULL, 2400 },
         {NULL,                    0,                 NULL,    0 }
       };
       static struct option service_long_options[] = {
@@ -712,7 +820,7 @@ main (int argc, char *argv[])
 	  }
 	  else
 	    {
-	      if (!opt.m_delete && check_for_pam_module ("pam_ldap.so", opt.force) != 0)
+	      if (!opt.m_delete && check_for_pam_module ("pam_ldap.so", opt.force, 0) != 0)
 		return 1;
 	      opt_set = mod_pam_ldap.get_opt_set (&mod_pam_ldap, ACCOUNT);
 	      opt_set->enable (opt_set, "is_enabled", opt.opt_val);
@@ -733,7 +841,7 @@ main (int argc, char *argv[])
 	    print_module_config (common_module_list, "pam_nam.so");
 	  else
 	    {
-	      if (!opt.m_delete && check_for_pam_module ("pam_nam.so", opt.force) != 0)
+	      if (!opt.m_delete && check_for_pam_module ("pam_nam.so", opt.force, 0) != 0)
 		return 1;
 
 	      opt_set = mod_pam_nam.get_opt_set (&mod_pam_nam, ACCOUNT);
@@ -755,7 +863,7 @@ main (int argc, char *argv[])
             print_module_config (common_module_list, "pam_winbind.so");
 	  else
 	    {
-	      if (!opt.m_delete && check_for_pam_module ("pam_winbind.so", opt.force) != 0)
+	      if (!opt.m_delete && check_for_pam_module ("pam_winbind.so", opt.force, 0) != 0)
 		return 1;
               opt_set = mod_pam_winbind.get_opt_set (&mod_pam_winbind,
 						     ACCOUNT);
@@ -780,7 +888,7 @@ main (int argc, char *argv[])
 	    print_module_config (common_module_list, "pam_sss.so");
 	  else
 	    {
-	      if (!opt.m_delete && check_for_pam_module ("pam_sss.so", opt.force) != 0)
+	      if (!opt.m_delete && check_for_pam_module ("pam_sss.so", opt.force, 0) != 0)
 		return 1;
 	      opt_set = mod_pam_sss.get_opt_set (&mod_pam_sss, ACCOUNT);
 	      opt_set->enable (opt_set, "is_enabled", opt.opt_val);
@@ -795,20 +903,6 @@ main (int argc, char *argv[])
 	      opt_set->enable (opt_set, "is_enabled", opt.opt_val);
 	    }
 	  break;
-	case 2400:
-	  /* pam_ecryptfs.so */
-	  if (opt.m_query)
-	    print_module_config (common_module_list, "pam_ecryptfs.so");
-	  else
-	    {
-	      if (!opt.m_delete && check_for_pam_module ("pam_ecryptfs.so", opt.force) != 0)
-		return 1;
-	      opt_set = mod_pam_ecryptfs.get_opt_set (&mod_pam_ecryptfs, AUTH);
-	      opt_set->enable (opt_set, "is_enabled", opt.opt_val);
-	      opt_set = mod_pam_ecryptfs.get_opt_set (&mod_pam_ecryptfs, SESSION);
-	      opt_set->enable (opt_set, "is_enabled", opt.opt_val);
-	    }
-	  break;
 	  /* From here we have single service modules */
 	case 3200:
 	  /* pam_cryptpass.so */
@@ -816,7 +910,7 @@ main (int argc, char *argv[])
 	    print_module_config (service_module_list, "pam_cryptpass.so");
 	  else
 	    {
-	      if (!opt.m_delete && check_for_pam_module ("pam_cryptpass.so", opt.force) != 0)
+	      if (!opt.m_delete && check_for_pam_module ("pam_cryptpass.so", opt.force, 0) != 0)
 		return 1;
 	      opt_set = mod_pam_cryptpass.get_opt_set (&mod_pam_cryptpass, SESSION);
 	      opt_set->enable (opt_set, "is_enabled", opt.opt_val);
@@ -828,7 +922,7 @@ main (int argc, char *argv[])
 	    print_module_config (service_module_list, "pam_cryptpass.so");
 	  else
 	    {
-	      if (!opt.m_delete && check_for_pam_module ("pam_cryptpass.so", opt.force) != 0)
+	      if (!opt.m_delete && check_for_pam_module ("pam_cryptpass.so", opt.force, 0) != 0)
 		return 1;
 	      opt_set = mod_pam_cryptpass.get_opt_set (&mod_pam_cryptpass, PASSWORD);
 	      opt_set->enable (opt_set, "is_enabled", opt.opt_val);
@@ -936,7 +1030,7 @@ main (int argc, char *argv[])
 	      opt_set_password->is_enabled (opt_set_password, "is_enabled") ||
 	      opt_set_session->is_enabled (opt_set_session, "is_enabled"))
 	    {
-	      if (check_for_pam_module ((*modptr)->name, 0))
+	      if (check_for_pam_module ((*modptr)->name, 0, 0))
 		retval = 1;
 	    }
 	  ++modptr;
@@ -960,9 +1054,11 @@ main (int argc, char *argv[])
       if (sanitize_check_auth (common_module_list, 0) != 0)
 	return 1;
 
-      opt_set = mod_pam_cracklib.get_opt_set (&mod_pam_cracklib, PASSWORD);
-      opt_set->enable (opt_set, "is_enabled", TRUE);
-
+      if (check_for_pam_module ("pam_pwquality.so", 0, 0) == 0)
+	{
+	  opt_set = mod_pam_pwquality.get_opt_set (&mod_pam_pwquality, PASSWORD);
+	  opt_set->enable (opt_set, "is_enabled", TRUE);
+	}
       opt_set = mod_pam_unix.get_opt_set (&mod_pam_unix, PASSWORD);
       opt_set->enable (opt_set, "is_enabled", TRUE);
       opt_set->enable (opt_set, "nullok", TRUE);
@@ -978,6 +1074,11 @@ main (int argc, char *argv[])
       opt_set->enable (opt_set, "is_enabled", opt.opt_val);
       opt_set = mod_pam_umask.get_opt_set (&mod_pam_umask, SESSION);
       opt_set->enable (opt_set, "is_enabled", opt.opt_val);
+      if (check_for_pam_module ("pam_systemd.so", 0, 1) == 0)
+	{
+	  opt_set = mod_pam_systemd.get_opt_set (&mod_pam_systemd, SESSION);
+	  opt_set->enable (opt_set, "is_enabled", TRUE);
+	}
       if (sanitize_check_session (common_module_list, 0) != 0)
 	return 1;
 
@@ -993,6 +1094,21 @@ main (int argc, char *argv[])
 
       if (write_config (confdir, CONF_SESSION_PC, SESSION, module_list_session) != 0)
 	return 1;
+
+      if (write_config (confdir, CONF_SESSION_NONLOGIN_PC, SESSION, module_list_session_nl) != 0)
+	return 1;
+
+      if (write_config (confdir, CONF_POSTLOGIN_ACCOUNT_PC, ACCOUNT, module_list_postlogin_account) != 0)
+	return 1;
+
+      if (write_config (confdir, CONF_POSTLOGIN_AUTH_PC, AUTH, module_list_postlogin_auth) != 0)
+	return 1;
+
+      if (write_config (confdir, CONF_POSTLOGIN_PASSWORD_PC, PASSWORD, module_list_postlogin_password) != 0)
+	return 1;
+
+      if (write_config (confdir, CONF_POSTLOGIN_SESSION_PC, SESSION, module_list_postlogin_session) != 0)
+	return 1;
     }
   else if (!gl_service)
     {
@@ -1024,6 +1140,21 @@ main (int argc, char *argv[])
 
       if (write_config (confdir, CONF_SESSION_PC, SESSION, module_list_session) != 0)
 	return 1;
+
+      if (write_config (confdir, CONF_SESSION_NONLOGIN_PC, SESSION, module_list_session_nl) != 0)
+	return 1;
+
+      if (write_config (confdir, CONF_POSTLOGIN_ACCOUNT_PC, ACCOUNT, module_list_postlogin_account) != 0)
+	return 1;
+
+      if (write_config (confdir, CONF_POSTLOGIN_AUTH_PC, AUTH, module_list_postlogin_auth) != 0)
+	return 1;
+
+      if (write_config (confdir, CONF_POSTLOGIN_PASSWORD_PC, PASSWORD, module_list_postlogin_password) != 0)
+	return 1;
+
+      if (write_config (confdir, CONF_POSTLOGIN_SESSION_PC, SESSION, module_list_postlogin_session) != 0)
+	return 1;
     }
   else
     {
@@ -1033,19 +1164,10 @@ main (int argc, char *argv[])
       if (debug)
 		  printf ("*** write_config (%s/pam.d/%s)\n", confdir, gl_service);
 
-      /* Check if service file exists */
-      char *conffile;
-      if (asprintf (&conffile, "%s/pam.d/%s", confdir, gl_service) < 0)
-	return 1;
-
-      if (access (conffile, R_OK) != 0)
-      {
-	fprintf (stderr, _("Cannot access '%s': %m\n"), conffile);
-	free (conffile);
-	return 1;
-      }
-      free (conffile);
-
+      /*
+       * Note that the modules in service_module_list[]
+       * do not use the "op" and the "fp" parameters.
+       */
       while (*modptr != NULL)
 	{
 	  retval |= (*modptr)->write_config (*modptr, -1, NULL);
@@ -1067,6 +1189,21 @@ main (int argc, char *argv[])
       if (relink (confdir, CONF_SESSION, CONF_SESSION_PC) != 0)
 	retval = 1;
 
+      if (relink (confdir, CONF_SESSION_NONLOGIN, CONF_SESSION_NONLOGIN_PC) != 0)
+	retval = 1;
+
+      if (relink (confdir, CONF_POSTLOGIN_ACCOUNT, CONF_POSTLOGIN_ACCOUNT_PC) != 0)
+	retval = 1;
+
+      if (relink (confdir, CONF_POSTLOGIN_AUTH, CONF_POSTLOGIN_AUTH_PC) != 0)
+	retval = 1;
+
+      if (relink (confdir, CONF_POSTLOGIN_PASSWORD, CONF_POSTLOGIN_PASSWORD_PC) != 0)
+	retval = 1;
+
+      if (relink (confdir, CONF_POSTLOGIN_SESSION, CONF_POSTLOGIN_SESSION_PC) != 0)
+	retval = 1;
+
       if (opt.m_init && retval == 0)
 	{
 	  rename ("/etc/security/pam_pwcheck.conf",
@@ -1089,6 +1226,21 @@ main (int argc, char *argv[])
 
       if (relink (confdir, CONF_SESSION, CONF_SESSION_PC) != 0)
 	retval = 1;
+
+      if (relink (confdir, CONF_SESSION_NONLOGIN, CONF_SESSION_NONLOGIN_PC) != 0)
+	retval = 1;
+
+      if (relink (confdir, CONF_POSTLOGIN_ACCOUNT, CONF_POSTLOGIN_ACCOUNT_PC) != 0)
+	retval = 1;
+
+      if (relink (confdir, CONF_POSTLOGIN_AUTH, CONF_POSTLOGIN_AUTH_PC) != 0)
+	retval = 1;
+
+      if (relink (confdir, CONF_POSTLOGIN_PASSWORD, CONF_POSTLOGIN_PASSWORD_PC) != 0)
+	retval = 1;
+
+      if (relink (confdir, CONF_POSTLOGIN_SESSION, CONF_POSTLOGIN_SESSION_PC) != 0)
+	retval = 1;
     }
 
   if (!gl_service)
@@ -1101,6 +1253,17 @@ main (int argc, char *argv[])
 	retval = 1;
       if (check_symlink (confdir, CONF_SESSION_PC, CONF_SESSION) != 0)
 	retval = 1;
+      if (check_symlink (confdir, CONF_SESSION_NONLOGIN_PC, CONF_SESSION_NONLOGIN) != 0)
+	retval = 1;
+
+      if (check_symlink (confdir, CONF_POSTLOGIN_ACCOUNT_PC, CONF_POSTLOGIN_ACCOUNT) != 0)
+	retval = 1;
+      if (check_symlink (confdir, CONF_POSTLOGIN_AUTH_PC, CONF_POSTLOGIN_AUTH) != 0)
+	retval = 1;
+      if (check_symlink (confdir, CONF_POSTLOGIN_PASSWORD_PC, CONF_POSTLOGIN_PASSWORD) != 0)
+	retval = 1;
+      if (check_symlink (confdir, CONF_POSTLOGIN_SESSION_PC, CONF_POSTLOGIN_SESSION) != 0)
+	retval = 1;
     }
 
   return retval;
